"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _1 = require("./");
class WorkflowExecute {
    constructor(additionalData, mode) {
        this.executionId = null;
        this.additionalData = additionalData;
        this.activeExecutions = _1.ActiveExecutions.getInstance();
        this.mode = mode;
    }
    async run(workflow, startNodes, destinationNode) {
        startNodes = startNodes || workflow.getStartNodes(destinationNode);
        let runNodeFilter = undefined;
        if (destinationNode) {
            runNodeFilter = workflow.getParentNodes(destinationNode);
            runNodeFilter.push(destinationNode);
        }
        const nodeExecutionStack = [];
        startNodes.forEach((node) => {
            nodeExecutionStack.push({
                node,
                data: {
                    main: [
                        [
                            {
                                json: {},
                            },
                        ],
                    ],
                },
            });
        });
        const runExecutionData = {
            startData: {
                destinationNode,
                runNodeFilter,
            },
            resultData: {
                runData: {},
            },
            executionData: {
                contextData: {},
                nodeExecutionStack,
                waitingExecution: {},
            },
        };
        return this.runExecutionData(workflow, runExecutionData);
    }
    async runPartialWorkflow(workflow, runData, startNodes, destinationNode) {
        let incomingNodeConnections;
        let connection;
        const runIndex = 0;
        const nodeExecutionStack = [];
        const waitingExecution = {};
        for (const startNode of startNodes) {
            incomingNodeConnections = workflow.connectionsByDestinationNode[startNode];
            const incomingData = [];
            if (incomingNodeConnections === undefined) {
                incomingData.push([
                    {
                        json: {}
                    }
                ]);
            }
            else {
                for (const connections of incomingNodeConnections.main) {
                    for (let inputIndex = 0; inputIndex < connections.length; inputIndex++) {
                        connection = connections[inputIndex];
                        incomingData.push(runData[connection.node][runIndex].data[connection.type][connection.index]);
                    }
                }
            }
            const executeData = {
                node: workflow.getNode(startNode),
                data: {
                    main: incomingData,
                }
            };
            nodeExecutionStack.push(executeData);
            incomingNodeConnections = workflow.connectionsByDestinationNode[destinationNode];
            if (incomingNodeConnections !== undefined) {
                for (const connections of incomingNodeConnections.main) {
                    for (let inputIndex = 0; inputIndex < connections.length; inputIndex++) {
                        connection = connections[inputIndex];
                        if (waitingExecution[destinationNode] === undefined) {
                            waitingExecution[destinationNode] = {};
                        }
                        if (waitingExecution[destinationNode][runIndex] === undefined) {
                            waitingExecution[destinationNode][runIndex] = {};
                        }
                        if (waitingExecution[destinationNode][runIndex][connection.type] === undefined) {
                            waitingExecution[destinationNode][runIndex][connection.type] = [];
                        }
                        if (runData[connection.node] !== undefined) {
                            waitingExecution[destinationNode][runIndex][connection.type].push(runData[connection.node][runIndex].data[connection.type][connection.index]);
                        }
                        else {
                            waitingExecution[destinationNode][runIndex][connection.type].push(null);
                        }
                    }
                }
            }
        }
        let runNodeFilter = undefined;
        runNodeFilter = workflow.getParentNodes(destinationNode);
        runNodeFilter.push(destinationNode);
        const runExecutionData = {
            startData: {
                destinationNode,
                runNodeFilter,
            },
            resultData: {
                runData,
            },
            executionData: {
                contextData: {},
                nodeExecutionStack,
                waitingExecution,
            },
        };
        return await this.runExecutionData(workflow, runExecutionData);
    }
    async executeHook(hookName, parameters) {
        if (this.additionalData.hooks === undefined) {
            return parameters[0];
        }
        if (this.additionalData.hooks[hookName] === undefined || this.additionalData.hooks[hookName].length === 0) {
            return parameters[0];
        }
        for (const hookFunction of this.additionalData.hooks[hookName]) {
            await hookFunction.apply(this, parameters)
                .catch((error) => {
                console.error(`There was a problem executing hook: "${hookName}"`);
                console.error('Parameters:');
                console.error(parameters);
                console.error('Error:');
                console.error(error);
            });
        }
    }
    addNodeToBeExecuted(workflow, runExecutionData, connectionData, outputIndex, parentNodeName, nodeSuccessData, runIndex) {
        let stillDataMissing = false;
        if (workflow.connectionsByDestinationNode[connectionData.node]['main'].length > 1) {
            let nodeWasWaiting = true;
            if (runExecutionData.executionData.waitingExecution[connectionData.node] === undefined) {
                runExecutionData.executionData.waitingExecution[connectionData.node] = {};
                nodeWasWaiting = false;
            }
            if (runExecutionData.executionData.waitingExecution[connectionData.node][runIndex] === undefined) {
                runExecutionData.executionData.waitingExecution[connectionData.node][runIndex] = {
                    main: []
                };
                for (let i = 0; i < workflow.connectionsByDestinationNode[connectionData.node]['main'].length; i++) {
                    runExecutionData.executionData.waitingExecution[connectionData.node][runIndex].main.push(null);
                }
            }
            if (nodeSuccessData === null) {
                runExecutionData.executionData.waitingExecution[connectionData.node][runIndex].main[connectionData.index] = null;
            }
            else {
                runExecutionData.executionData.waitingExecution[connectionData.node][runIndex].main[connectionData.index] = nodeSuccessData[outputIndex];
            }
            let thisExecutionData;
            let allDataFound = true;
            for (let i = 0; i < runExecutionData.executionData.waitingExecution[connectionData.node][runIndex].main.length; i++) {
                thisExecutionData = runExecutionData.executionData.waitingExecution[connectionData.node][runIndex].main[i];
                if (thisExecutionData === null) {
                    allDataFound = false;
                    break;
                }
            }
            if (allDataFound === true) {
                runExecutionData.executionData.nodeExecutionStack.push({
                    node: workflow.nodes[connectionData.node],
                    data: runExecutionData.executionData.waitingExecution[connectionData.node][runIndex]
                });
                delete runExecutionData.executionData.waitingExecution[connectionData.node][runIndex];
                if (Object.keys(runExecutionData.executionData.waitingExecution[connectionData.node]).length === 0) {
                    delete runExecutionData.executionData.waitingExecution[connectionData.node];
                }
                return;
            }
            else {
                stillDataMissing = true;
            }
            if (nodeWasWaiting === false) {
                const checkOutputNodes = [];
                for (const outputIndexParent in workflow.connectionsBySourceNode[parentNodeName].main) {
                    if (!workflow.connectionsBySourceNode[parentNodeName].main.hasOwnProperty(outputIndexParent)) {
                        continue;
                    }
                    for (const connectionDataCheck of workflow.connectionsBySourceNode[parentNodeName].main[outputIndexParent]) {
                        checkOutputNodes.push(connectionDataCheck.node);
                    }
                }
                for (let inputIndex = 0; inputIndex < workflow.connectionsByDestinationNode[connectionData.node]['main'].length; inputIndex++) {
                    for (const inputData of workflow.connectionsByDestinationNode[connectionData.node]['main'][inputIndex]) {
                        if (inputData.node === parentNodeName) {
                            continue;
                        }
                        const executionStackNodes = runExecutionData.executionData.nodeExecutionStack.map((stackData) => stackData.node.name);
                        if (inputData.node !== parentNodeName && checkOutputNodes.includes(inputData.node)) {
                            continue;
                        }
                        if (executionStackNodes.includes(inputData.node)) {
                            continue;
                        }
                        if (runExecutionData.resultData.runData[inputData.node] !== undefined) {
                            continue;
                        }
                        const parentNodes = workflow.getParentNodes(inputData.node, 'main', -1);
                        let nodeToAdd = inputData.node;
                        parentNodes.push(inputData.node);
                        parentNodes.reverse();
                        for (const parentNode of parentNodes) {
                            if (inputData.node !== parentNode && checkOutputNodes.includes(parentNode)) {
                                nodeToAdd = undefined;
                                break;
                            }
                            if (executionStackNodes.includes(parentNode)) {
                                nodeToAdd = undefined;
                                break;
                            }
                            if (runExecutionData.resultData.runData[parentNode] !== undefined) {
                                break;
                            }
                            nodeToAdd = parentNode;
                        }
                        if (nodeToAdd === undefined) {
                            continue;
                        }
                        if (workflow.connectionsByDestinationNode[nodeToAdd] === undefined) {
                            runExecutionData.executionData.nodeExecutionStack.push({
                                node: workflow.getNode(nodeToAdd),
                                data: {
                                    main: [
                                        [
                                            {
                                                json: {},
                                            },
                                        ],
                                    ],
                                },
                            });
                        }
                    }
                }
            }
        }
        const connectionDataArray = [];
        for (let i = connectionData.index; i >= 0; i--) {
            connectionDataArray[i] = null;
        }
        if (nodeSuccessData === null) {
            connectionDataArray[connectionData.index] = null;
        }
        else {
            connectionDataArray[connectionData.index] = nodeSuccessData[outputIndex];
        }
        if (stillDataMissing === true) {
            if (!runExecutionData.executionData.waitingExecution.hasOwnProperty(connectionData.node)) {
                runExecutionData.executionData.waitingExecution[connectionData.node] = {};
            }
            runExecutionData.executionData.waitingExecution[connectionData.node][runIndex] = {
                main: connectionDataArray
            };
        }
        else {
            runExecutionData.executionData.nodeExecutionStack.push({
                node: workflow.nodes[connectionData.node],
                data: {
                    main: connectionDataArray
                }
            });
        }
    }
    async runExecutionData(workflow, runExecutionData) {
        const startedAt = new Date().getTime();
        const workflowIssues = workflow.checkReadyForExecution();
        if (workflowIssues !== null) {
            throw new Error('The workflow has issues and can for that reason not be executed. Please fix them first.');
        }
        let executionData;
        let executionError;
        let executionNode;
        let nodeSuccessData;
        let runIndex;
        let startTime;
        let taskData;
        if (runExecutionData.startData === undefined) {
            runExecutionData.startData = {};
        }
        this.executionId = this.activeExecutions.add(workflow, runExecutionData, this.mode);
        this.executeHook('workflowExecuteBefore', [this.executionId]);
        let currentExecutionTry = '';
        let lastExecutionTry = '';
        process.nextTick(() => (async () => {
            executionLoop: while (runExecutionData.executionData.nodeExecutionStack.length !== 0) {
                if (this.activeExecutions.shouldBeStopped(this.executionId) === true) {
                    break;
                }
                nodeSuccessData = null;
                executionError = undefined;
                executionData = runExecutionData.executionData.nodeExecutionStack.shift();
                executionNode = executionData.node;
                this.executeHook('nodeExecuteBefore', [this.executionId, executionNode.name]);
                runIndex = 0;
                if (runExecutionData.resultData.runData.hasOwnProperty(executionNode.name)) {
                    runIndex = runExecutionData.resultData.runData[executionNode.name].length;
                }
                currentExecutionTry = `${executionNode.name}:${runIndex}`;
                if (currentExecutionTry === lastExecutionTry) {
                    throw new Error('Did stop execution because execution seems to be in endless loop.');
                }
                if (runExecutionData.startData.runNodeFilter !== undefined && runExecutionData.startData.runNodeFilter.indexOf(executionNode.name) === -1) {
                    continue;
                }
                if (workflow.connectionsByDestinationNode.hasOwnProperty(executionNode.name)) {
                    if (workflow.connectionsByDestinationNode[executionNode.name].hasOwnProperty('main')) {
                        let inputConnections;
                        let connectionIndex;
                        inputConnections = workflow.connectionsByDestinationNode[executionNode.name]['main'];
                        for (connectionIndex = 0; connectionIndex < inputConnections.length; connectionIndex++) {
                            if (workflow.getHighestNode(executionNode.name, 'main', connectionIndex).length === 0) {
                                continue;
                            }
                            if (!executionData.data.hasOwnProperty('main')) {
                                runExecutionData.executionData.nodeExecutionStack.push(executionData);
                                lastExecutionTry = currentExecutionTry;
                                continue executionLoop;
                            }
                            if (executionData.data.main.length < connectionIndex || executionData.data.main[connectionIndex] === null) {
                                runExecutionData.executionData.nodeExecutionStack.push(executionData);
                                lastExecutionTry = currentExecutionTry;
                                continue executionLoop;
                            }
                        }
                    }
                }
                startTime = new Date().getTime();
                let maxTries = 1;
                if (executionData.node.retryOnFail === true) {
                    maxTries = Math.min(5, Math.max(2, executionData.node.maxTries || 3));
                }
                let waitBetweenTries = 0;
                if (executionData.node.retryOnFail === true) {
                    waitBetweenTries = Math.min(5000, Math.max(0, executionData.node.waitBetweenTries || 1000));
                }
                for (let tryIndex = 0; tryIndex < maxTries; tryIndex++) {
                    try {
                        if (tryIndex !== 0) {
                            executionError = undefined;
                            if (waitBetweenTries !== 0) {
                                await new Promise((resolve) => {
                                    setTimeout(() => {
                                        resolve();
                                    }, waitBetweenTries);
                                });
                            }
                        }
                        if (this.activeExecutions.shouldBeStopped(this.executionId) === true) {
                            break;
                        }
                        runExecutionData.resultData.lastNodeExecuted = executionData.node.name;
                        nodeSuccessData = await workflow.runNode(executionData.node, JSON.parse(JSON.stringify(executionData.data)), runExecutionData, runIndex, this.additionalData, _1.NodeExecuteFunctions, this.mode);
                        if (nodeSuccessData === null) {
                            continue executionLoop;
                        }
                        break;
                    }
                    catch (error) {
                        executionError = {
                            message: error.message,
                            stack: error.stack,
                        };
                    }
                }
                if (!runExecutionData.resultData.runData.hasOwnProperty(executionNode.name)) {
                    runExecutionData.resultData.runData[executionNode.name] = [];
                }
                taskData = {
                    startTime,
                    executionTime: (new Date().getTime()) - startTime
                };
                if (executionError !== undefined) {
                    taskData.error = executionError;
                    if (executionData.node.continueOnFail === true) {
                        if (executionData.data.hasOwnProperty('main') && executionData.data.main.length > 0) {
                            if (executionData.data.main[0] !== null) {
                                nodeSuccessData = [JSON.parse(JSON.stringify(executionData.data.main[0]))];
                            }
                        }
                    }
                    else {
                        runExecutionData.resultData.runData[executionNode.name].push(taskData);
                        runExecutionData.executionData.nodeExecutionStack.unshift(executionData);
                        this.executeHook('nodeExecuteAfter', [this.executionId, executionNode.name, taskData]);
                        break;
                    }
                }
                taskData.data = {
                    'main': nodeSuccessData
                };
                this.executeHook('nodeExecuteAfter', [this.executionId, executionNode.name, taskData]);
                runExecutionData.resultData.runData[executionNode.name].push(taskData);
                if (runExecutionData.startData && runExecutionData.startData.destinationNode && runExecutionData.startData.destinationNode === executionNode.name) {
                    continue;
                }
                if (workflow.connectionsBySourceNode.hasOwnProperty(executionNode.name)) {
                    if (workflow.connectionsBySourceNode[executionNode.name].hasOwnProperty('main')) {
                        let outputIndex, connectionData;
                        for (outputIndex in workflow.connectionsBySourceNode[executionNode.name]['main']) {
                            if (!workflow.connectionsBySourceNode[executionNode.name]['main'].hasOwnProperty(outputIndex)) {
                                continue;
                            }
                            for (connectionData of workflow.connectionsBySourceNode[executionNode.name]['main'][outputIndex]) {
                                if (!workflow.nodes.hasOwnProperty(connectionData.node)) {
                                    return Promise.reject(new Error(`The node "${executionNode.name}" connects to not found node "${connectionData.node}"`));
                                }
                                this.addNodeToBeExecuted(workflow, runExecutionData, connectionData, parseInt(outputIndex, 10), executionNode.name, nodeSuccessData, runIndex);
                            }
                        }
                    }
                }
            }
            return Promise.resolve();
        })()
            .then(async () => {
            const fullRunData = {
                data: runExecutionData,
                mode: this.mode,
                startedAt: new Date(startedAt),
                stoppedAt: new Date(),
            };
            if (executionError !== undefined) {
                fullRunData.data.resultData.error = executionError;
            }
            else {
                fullRunData.finished = true;
            }
            this.activeExecutions.remove(this.executionId, fullRunData);
            await this.executeHook('workflowExecuteAfter', [fullRunData, this.executionId]);
            return fullRunData;
        })
            .catch(async (error) => {
            const fullRunData = {
                data: runExecutionData,
                mode: this.mode,
                startedAt: new Date(startedAt),
                stoppedAt: new Date(),
            };
            fullRunData.data.resultData.error = {
                message: error.message,
                stack: error.stack,
            };
            this.activeExecutions.remove(this.executionId, fullRunData);
            await this.executeHook('workflowExecuteAfter', [fullRunData, this.executionId]);
            return fullRunData;
        }));
        return this.executionId;
    }
}
exports.WorkflowExecute = WorkflowExecute;
//# sourceMappingURL=WorkflowExecute.js.map